package com.practice;

/*
 * jcg : category: core java
 * 1. brute force method
 * 2. sliding window method
 * 3. sliding window optimzied
 */

class Bruteforce {

  // find all substrings of a string in a loop from start till end.
  // to check if substring contains all unique characters put all characters in set one by one.
  // if any character is present in set, skip that string else consider its length.
  public long bruteforce(String str) {
    int n = str.length();
    int res = 0;

    for (int i = 0; i < n; i++) {
      for (int j = i; j < n; j++) {
        if (checkRepetition(str, i, j)) {
          res = Math.max(res, j - i + 1);
        }
      }
    }
    return res;
  }

  private boolean checkRepetition(String s, int start, int end) {
    int[] chars = new int[128];

    for (int i = start; i <= end; i++) {
      char c = s.charAt(i);
      chars[c]++;
      if (chars[c] > 1) {
        return false;
      }
    }
    return true;
  }
}

class SlidingWindow {

  // in this approach we will skip the need of considering a substring and check if contains a
  // duplicate character. we will focus on improving the repeated tasks i.e. if the substring
  // from index to j-1 has already been checked to have no duplicate characters then simplify the
  // check if s[j] is a substring of s[i][j].
  public long slidingwindow(String str) {
    int n = str.length();
    int res = 0;

    for (int i = 0; i < n; i++) {
      boolean[] visited = new boolean[256];
      for (int j = i; j < n; j++) {
        if (visited[str.charAt(j)]) {
          break;
        } else {
          res = Math.max(res, j - i + 1);
          visited[str.charAt(j)] = true;
        }
      }
      visited[str.charAt(i)] = false;
    }
    return res;
  }
}

class OptimizedSlidingWindow {

  // in the sliding approach checking if substring takes another string the complexity is o(n^2)
  // time. to improve it further we can use hashset as a sliding window. here we check if a
  // character has already been visited and is present in current window it can be done in o(1).
  // here we will use left and right pointers to keep a track of maximum non-repating substring
  // in a string.
  public long optimizedslidingwindow(String str) {
    int[] chars = new int[128];
    int left = 0;
    int right = 0;
    int res = 0;

    while (right < str.length()) {
      char r = str.charAt(right);
      chars[r]++;
      while (chars[r] > 1) {
        char l = str.charAt(left);
        chars[l]--;
        left++;
      }
      res = Math.max(res, right - left + 1);
      right++;
    }
    return res;
  }
}

public class LongestStringWithoutRepeatingCharactersEx {

  public static void main(String[] args) {
    Bruteforce bf = new Bruteforce();
    System.out.println("\nBruteforce method = " + bf.bruteforce("enjoyalgorithms"));

    SlidingWindow sd = new SlidingWindow();
    System.out.println("\nSlidingwindow method = " + sd.slidingwindow("enjoyalgorithms"));

    OptimizedSlidingWindow osd = new OptimizedSlidingWindow();
    System.out.println(
        "\nOptimizedslidingwindow method = " + osd.optimizedslidingwindow("enjoyalgorithms"));
  }
}
